diff -r -u ../asterisk.orig/apps/app_dial.c ./apps/app_dial.c
--- ../asterisk.orig/apps/app_dial.c	2003-03-07 14:25:36.000000000 +0300
+++ ./apps/app_dial.c	2003-03-07 14:28:05.000000000 +0300
@@ -35,6 +35,8 @@
 
 #include <pthread.h>
 
+#define AST_RECORD_DIR	INSTALL_PREFIX "/var/spool/asterisk/recording"
+
 static char *tdesc = "Dialing Application";
 
 static char *app = "Dial";
@@ -64,6 +66,15 @@
 "      'm' -- provide hold music to the calling party until answered.\n"
 "      'd' -- data-quality (modem) call (minimum delay).\n"
 "      'H' -- allow caller to hang up by hitting *.\n"
+"      'R' -- force recording sound on the briged call.\n"
+"	      files will be written to the /var/spool/asterisk/recording/.\n"
+"	      Name format: chan0_chan1_callingid_calledid_timestamp_x.pcm\n"
+"	      where x - 'i' for one peer and 'o' for other peer."
+"	      Recording performed without conferencing to minimize owerload."
+"      'D' -- Interpret URL as DTMF digits and send them to the dialed peer.\n"
+"	      If DTMF digit string begins with 'e' than appiarence of '99' will be\n"
+"	      replaced with single '9' and other combinations of '9X' will be\n"
+"	      treated as pause in 1 second when playing DTMF.\n"
 "      'P[(x)]' -- privacy mode, using 'x' as database if provided.\n"
 "  In addition to transferring the call, a call may be parked and then picked\n"
 "up by another user.\n"
@@ -82,6 +93,7 @@
 	int musiconhold;
 	int dataquality;
 	int allowdisconnect;
+	int force_record;
 	struct localuser *next;
 };
 
@@ -162,6 +174,10 @@
 			} else {
 				if (option_verbose > 2)
 					ast_verbose( VERBOSE_PREFIX_2 "No one is available to answer at this time\n");
+				/* See if there is a special busy message */
+				/* I think that we MUST indicate this situation as BUSY */
+				if (ast_exists_extension(in, in->context, in->exten, in->priority + 101, in->callerid)) 
+					in->priority+=100;
 			}
 			*to = 0;
 			/* if no one available we'd better stop MOH/ringing to */
@@ -301,6 +317,10 @@
 	char *newnum;
 	char callerid[256], *l, *n;
 	char *url=NULL; /* JDG */
+	char fname[256];
+	time_t st_time;
+	struct tm *ts;
+	int url_as_dtmf;
 	struct ast_var_t *current;
 	struct varshead *headp, *newheadp;
 	struct ast_var_t *newvar;
@@ -420,6 +440,10 @@
                         else    tmp->dataquality = 0;
 			if (strchr(transfer, 'H'))
 				tmp->allowdisconnect = 1;
+			if (strchr(transfer, 'R'))
+				tmp->force_record = 1;
+			if (strchr(transfer, 'D'))
+				url_as_dtmf = 1;
                         else    tmp->allowdisconnect = 0;
 		}
 		strncpy(numsubst, number, sizeof(numsubst)-1);
@@ -488,6 +512,7 @@
 			tmp->chan->ani = NULL;
 		/* Presense of ADSI CPE on outgoing channel follows ours */
 		tmp->chan->adsicpe = chan->adsicpe;
+		tmp->chan->force_record = tmp->force_record;
 		/* Place the call, but don't wait on the answer */
 		res = ast_call(tmp->chan, numsubst, 0);
 		if (res) {
@@ -559,11 +584,136 @@
 			return -1;
 		}
  		/* JDG: sendurl */
- 		if( url && strlen(url) && ast_channel_supports_html(peer) ) {
+ 		if( url && strlen(url) && ast_channel_supports_html(peer) && !url_as_dtmf) {
  			ast_log(LOG_DEBUG, "app_dial: sendurl=%s.\n", url);
  			ast_channel_sendurl( peer, url );
  		} /* /JDG */
+		if (url && strlen(url) && url_as_dtmf)
+		{
+			struct ast_frame f, *tf;
+			int x, dms = 250;
+			int res = 0, compat=1;
+			struct ast_channel *cs[2], *who;
+
+			if (((chan->writeformat != peer->readformat) || (chan->readformat != peer->writeformat)) &&
+			    !(chan->generator || peer->generator))
+				if (ast_channel_make_compatible(chan, peer))
+					compat = 0;
+			if (compat)
+			{
+				if (*url == 'e')
+				{
+					url++;
+					for (x=0;x<strlen(url);x++)
+					{
+						if ( x+1 < strlen(url) )
+						{
+							if (url[x] == '9')
+							{
+								if (url[x+1] != '9')
+									url[x] = ',';
+								strcpy(&url[x+1], &url[x+2]);
+							}
+						}
+					}
+				}
+				
+				if (option_verbose > 2)
+					ast_verbose(VERBOSE_PREFIX_3  "app_dial: sendtmf=%s.\n", url);
+				if ( !(res = ast_safe_sleep(peer, 1000)) )
+				{
+					cs[0] = chan;
+					cs[1] = peer;
+					for (x=0;x<strlen(url);x++)
+					{
+						if ((url[x] < '0' || url[x] > '9') && 
+						    (url[x] < 'A' || url[x] > 'D') &&
+						    (url[x] < 'a' || url[x] > 'd') &&
+						    url[x] != '#' && url[x] != '*' &&
+						    url[x] != ',')
+							continue;
+						if ( url[x] != ',' )
+						{
+							memset(&f, 0, sizeof(f));
+							f.frametype = AST_FRAME_DTMF;
+							f.subclass = url[x];
+							f.src = "app_dial";
+							if ( (res = ast_write(peer, &f)) )
+								break;
+							dms = 250;
+						}
+						else
+							dms = 1000;
+						while (dms > 0)
+						{
+							if ( !(who = ast_waitfor_n(cs, 2, &dms)) )
+								continue;
+							if ( !(tf = ast_read(who)) )
+							{
+								res = -1;
+								break;
+							}
+							if ( who == peer && 
+							      (tf->frametype == AST_FRAME_VOICE ||
+							       tf->frametype == AST_FRAME_TEXT ||
+							       tf->frametype == AST_FRAME_VIDEO ||
+							       tf->frametype == AST_FRAME_IMAGE ||
+							       tf->frametype == AST_FRAME_DTMF) )
+								res = ast_write(chan, tf);
+							ast_frfree(tf);
+							if (res)
+								break;
+						}
+						if (res)
+							break;
+					}
+					if (!res)
+						if (option_verbose > 2)
+							ast_verbose(VERBOSE_PREFIX_3 "Sent digit string '%s' on %s\n", url, peer->name);
+				}
+			}
+		}
+		if (peer->force_record)
+		{
+			char	chname0[256], chname1[256], *p;
+			/* Try to open streams for recording */
+			time(&st_time);
+			ts = localtime(&st_time);
+			strncpy(chname0,chan->name,sizeof(chname0));
+			strncpy(chname1,peer->name,sizeof(chname0));
+			while ( (p = strchr(chname0,'/')) )
+				*p = '-';
+			while ( (p = strchr(chname1,'/')) )
+				*p = '-';
+			snprintf(fname,sizeof(fname),"%s/%s_%s_%s_%s_%04d%02d%02d%02d%02d%02d_i",
+					AST_RECORD_DIR,chname0,chname1,peer->callerid,chan->exten,ts->tm_year+1900,
+					ts->tm_mon+1,ts->tm_mday,ts->tm_hour,ts->tm_min,ts->tm_sec);
+			if (option_verbose > 2)
+				ast_verbose(VERBOSE_PREFIX_3 "Force recording in voice on %s to %s.gsm\n",peer->name,fname);
+			if ( !(peer->in_stream = ast_writefile( fname, "gsm", NULL, O_CREAT|O_TRUNC|O_WRONLY , 0, 0644)) )
+				ast_log(LOG_WARNING, "Could not create file %s.gsm\n", fname);
+			snprintf(fname,sizeof(fname),"%s/%s_%s_%s_%s_%04d%02d%02d%02d%02d%02d_o",
+					AST_RECORD_DIR,chname0,chname1,peer->callerid,chan->exten,ts->tm_year+1900,
+					ts->tm_mon+1,ts->tm_mday,ts->tm_hour,ts->tm_min,ts->tm_sec);
+			if (option_verbose > 2)
+				ast_verbose(VERBOSE_PREFIX_3 "Force recording out voice on %s to %s.gsm\n",peer->name,fname);
+			if ( !(peer->out_stream = ast_writefile( fname, "gsm", NULL, O_CREAT|O_TRUNC|O_WRONLY , 0, 0644)) )
+				ast_log(LOG_WARNING, "Could not create file %s.gsm\n", fname);
+		}
 		res = ast_bridge_call(chan, peer, allowredir, allowdisconnect);
+		if (peer->force_record)
+		{
+			if (peer->in_stream)
+			{
+				ast_closestream(peer->in_stream);
+				peer->in_stream = NULL;
+			}
+			if (peer->out_stream)
+			{
+				ast_closestream(peer->out_stream);
+				peer->out_stream = NULL;
+			}
+		}
 		ast_hangup(peer);
 	}	
 out:
diff -r -u ../asterisk.orig/apps/app_queue.c ./apps/app_queue.c
--- ../asterisk.orig/apps/app_queue.c	2003-01-21 17:01:04.000000000 +0300
+++ ./apps/app_queue.c	2003-03-06 16:45:15.000000000 +0300
@@ -58,6 +58,7 @@
 "      'T' -- to allow the calling user to transfer the call.\n"
 "      'd' -- data-quality (modem) call (minimum delay).\n"
 "      'H' -- allow caller to hang up by hitting *.\n"
+"      'M' -- treat announceoverride as mandatory announcement.\n"
 "  In addition to transferring the call, a call may be parked and then picked\n"
 "up by another user.\n"
 "  The optionnal URL will be sent to the called party if the channel supports\n"
@@ -85,6 +86,9 @@
 	char moh[80];				/* Name of musiconhold to be used */
 	char announce[80];		/* Announcement to play */
 	char context[80];		/* Context when user exits queue */
+	char userannounce[80];		/* Announcement to play to calling user */
+	char mandatoryannounce[80];	/* Mandatory announcement for the calling user */
+	int useranntime;			/* Last time user announce was played */
 	int pos;					/* Where we are in the queue */
 	time_t start;				/* When we started holding */
 	struct ast_channel *chan;	/* Our channel */
@@ -103,7 +107,10 @@
 	char moh[80];			/* Name of musiconhold to be used */
 	char announce[80];		/* Announcement to play */
 	char context[80];		/* Announcement to play */
-	int announcetimeout;	/* How often to announce their position */
+	char userannounce[80];		/* Announcement to play to calling user */
+	char mandatoryannounce[80];	/* Mandatory announcement for the calling user */
+	int userannouncetimeout;	/* How often to play announce to calling user */
+	int announcetimeout;		/* How often to announce their position */
 	int count;				/* How many entries are in the queue */
 	int maxlen;				/* Max number of entries in queue */
 
@@ -150,8 +157,11 @@
 				qe->next = NULL;
 				qe->parent = q;
 				qe->pos = ++pos;
+				qe->useranntime = time(NULL);
 				strncpy(qe->moh, q->moh, sizeof(qe->moh));
 				strncpy(qe->announce, q->announce, sizeof(qe->announce));
+				strncpy(qe->mandatoryannounce, q->mandatoryannounce, sizeof(qe->mandatoryannounce));
+				strncpy(qe->userannounce, q->userannounce, sizeof(qe->userannounce));
 				strncpy(qe->context, q->context, sizeof(qe->context));
 				q->count++;
 				res = 0;
@@ -408,7 +418,7 @@
 static int try_calling(struct queue_ent *qe, char *options, char *announceoverride, char *url)
 {
 	struct member *cur;
-	struct localuser *outgoing=NULL, *tmp = NULL;
+	struct localuser *outgoing=NULL, *tmp = NULL, o;
 	int to;
 	int allowredir=0;
 	int allowdisconnect=0;
@@ -423,7 +433,7 @@
 	cur = qe->parent->members;
 	if (strlen(qe->announce))
 		announce = qe->announce;
-	if (announceoverride && strlen(announceoverride))
+	if (announceoverride && strlen(announceoverride) && (options && !strchr(options,'M')) )
 		announce = announceoverride;
 	while(cur) {
 		/* Get a technology/[device:]number pair */
@@ -610,11 +620,25 @@
 
 static int wait_a_bit(struct queue_ent *qe)
 {
-	int retrywait;
+	int retrywait, uatime, uatimeout, res;
 	/* Hold the lock while we setup the outgoing calls */
 	ast_pthread_mutex_lock(&qe->parent->lock);
 	retrywait = qe->parent->retry * 1000;
+	uatime = qe->useranntime;
+	uatimeout = qe->parent->userannouncetimeout;
+	if (uatime + uatimeout < time(NULL))
+		qe->useranntime = time(NULL);
 	ast_pthread_mutex_unlock(&qe->parent->lock);
+	if (*qe->userannounce && uatime + uatimeout < time(NULL))
+	{
+		ast_moh_stop(qe->chan);
+		res = ast_streamfile(qe->chan, qe->userannounce, qe->chan->language);
+		if (!res)
+			res = ast_waitstream(qe->chan, "");
+		if (!res)
+			ast_moh_start(qe->chan, qe->moh);
+		return res;
+	}
 	return ast_waitfordigit(qe->chan, retrywait);
 }
 
@@ -682,43 +706,61 @@
 	qe.start = time(NULL);
 	if (!join_queue(queuename, &qe)) {
 		/* Start music on hold */
-		ast_moh_start(chan, qe.moh);
-		for (;;) {
-			res = wait_our_turn(&qe);
-			/* If they hungup, return immediately */
-			if (res < 0) {
-				if (option_verbose > 2) {
-					ast_verbose(VERBOSE_PREFIX_3 "User disconnected while waiting their turn\n");
-					res = -1;
-				}
-				break;
-			}
+		if (strchr(options,'M') && announceoverride && *announceoverride)
+		{
+			res = ast_streamfile(chan, announceoverride, chan->language);
 			if (!res)
-				break;
-			if (valid_exit(&qe, res))
-				break;
+				res = ast_waitstream(chan, "");
 		}
-		if (!res) {
+		else if ( *qe.mandatoryannounce )
+		{
+			res = ast_streamfile(chan, qe.mandatoryannounce, chan->language);
+			if (!res)
+				res = ast_waitstream(chan, "");
+		}
+		else
+			res = 0;
+		qe.useranntime = time(NULL);
+		if (!res)
+		{
+			ast_moh_start(chan, qe.moh);
 			for (;;) {
-				res = try_calling(&qe, options, announceoverride, url);
-				if (res)
-					break;
-				res = wait_a_bit(&qe);
+				res = wait_our_turn(&qe);
+				/* If they hungup, return immediately */
 				if (res < 0) {
 					if (option_verbose > 2) {
-						ast_verbose(VERBOSE_PREFIX_3 "User disconnected when they almost made it\n");
+						ast_verbose(VERBOSE_PREFIX_3 "User disconnected while waiting their turn\n");
 						res = -1;
 					}
 					break;
 				}
-				if (res && valid_exit(&qe, res))
+				if (!res)
+					break;
+				if (valid_exit(&qe, res))
 					break;
 			}
+			if (!res) {
+				for (;;) {
+					res = try_calling(&qe, options, announceoverride, url);
+					if (res)
+						break;
+					res = wait_a_bit(&qe);
+					if (res < 0) {
+						if (option_verbose > 2) {
+							ast_verbose(VERBOSE_PREFIX_3 "User disconnected when they almost made it\n");
+							res = -1;
+						}
+						break;
+					}
+					if (res && valid_exit(&qe, res))
+						break;
+				}
+			}
+			ast_moh_stop(chan);
 		}
 		/* Don't allow return code > 0 */
 		if (res > 0)
 			res = 0;	
-		ast_moh_stop(chan);
 		leave_queue(&qe);
 	} else {
 		ast_log(LOG_WARNING, "Unable to join queue '%s'\n", queuename);
@@ -778,10 +820,13 @@
 				q->retry = 0;
 				q->timeout = -1;
 				q->maxlen = 0;
+				q->userannouncetimeout = 0;
 				free_members(q);
 				strcpy(q->moh, "");
 				strcpy(q->announce, "");
 				strcpy(q->context, "");
+				strcpy(q->userannounce, "");
+				strcpy(q->mandatoryannounce, "");
 				prev = NULL;
 				var = ast_variable_browse(cfg, cat);
 				while(var) {
@@ -816,6 +861,12 @@
 						q->retry = atoi(var->value);
 					} else if (!strcasecmp(var->name, "maxlen")) {
 						q->maxlen = atoi(var->value);
+					} else if (!strcasecmp(var->name, "mandatory")) {
+						strncpy(q->mandatoryannounce, var->value, sizeof(q->mandatoryannounce) - 1);
+					} else if (!strcasecmp(var->name, "periodic_announce")) {
+						strncpy(q->userannounce, var->value, sizeof(q->userannounce) - 1);
+					} else if (!strcasecmp(var->name, "periodic_timeout")) {
+						q->userannouncetimeout = atoi(var->value);
 					} else {
 						ast_log(LOG_WARNING, "Unknown keyword in queue '%s': %s at line %d of queue.conf\n", cat, var->name, var->lineno);
 					}
@@ -827,6 +878,8 @@
 					q->timeout = DEFAULT_TIMEOUT;
 				if (q->maxlen < 0)
 					q->maxlen = 0;
+				if (q->userannouncetimeout <= 0)
+					q->userannouncetimeout = 300;
 				if (!new) 
 					ast_pthread_mutex_unlock(&q->lock);
 				if (new) {
diff -r -u ../asterisk.orig/cdr/cdr_csv.c ./cdr/cdr_csv.c
--- ../asterisk.orig/cdr/cdr_csv.c	2003-03-07 14:25:39.000000000 +0300
+++ ./cdr/cdr_csv.c	2003-02-25 19:06:34.000000000 +0300
@@ -123,6 +123,10 @@
 
 static int build_csv_record(char *buf, int len, struct ast_cdr *cdr)
 {
+	struct ast_var_t *newvariable;
+        struct varshead *headp;
+	char	tmp[512];
+        
 	buf[0] = '\0';
 	/* Account code */
 	append_string(buf, cdr->accountcode, len);
@@ -156,6 +160,16 @@
 	append_string(buf, ast_cdr_disp2str(cdr->disposition), len);
 	/* AMA Flags */
 	append_string(buf, ast_cdr_flags2str(cdr->amaflags), len);
+	
+        headp=&cdr->parent_chan->varshead;
+                
+	AST_LIST_TRAVERSE (headp,newvariable,entries) {
+		if (strcasecmp(ast_var_name(newvariable),"tempvar"))
+		{
+			snprintf(tmp,sizeof(tmp)-1,"%s='%s'",ast_var_name(newvariable),ast_var_value(newvariable));
+			append_string(buf,tmp,len);
+		}
+	} 
 
 	/* If we hit the end of our buffer, log an error */
 	if (strlen(buf) < len - 5) {
@@ -188,7 +202,7 @@
 static int csv_log(struct ast_cdr *cdr)
 {
 	/* Make sure we have a big enough buf */
-	char buf[1024];
+	char buf[4096];
 	char csvmaster[AST_CONFIG_MAX_PATH];
 	snprintf((char *)csvmaster,sizeof(csvmaster)-1,"%s/%s/%s",(char *)ast_config_AST_LOG_DIR,CSV_LOG_DIR,CSV_MASTER);
 #if 0
diff -r -u ../asterisk.orig/cdr/cdr_mysql.c ./cdr/cdr_mysql.c
--- ../asterisk.orig/cdr/cdr_mysql.c	2002-11-25 22:46:01.000000000 +0300
+++ ./cdr/cdr_mysql.c	2003-02-20 18:05:03.000000000 +0300
@@ -28,7 +28,7 @@
 
 #include <mysql.h>
 
-#define DATE_FORMAT "%Y-%m-%d %T"
+#define DATE_FORMAT "%Y%m%d%H%M%S"
 
 static char *desc = "MySQL CDR Backend";
 static char *name = "mysql";
@@ -38,38 +38,48 @@
 
 static int mysql_log(struct ast_cdr *cdr)
 {
-  struct tm *tm;
-  struct timeval *tv;
-  struct timezone *tz;
-  char *sqlcmd, *timestr;
-  time_t t;
-
-
-  tv = (struct timeval *)malloc(sizeof(struct timeval));
-  tz = (struct timezone *)malloc(sizeof(struct timezone));
-  sqlcmd = (char *)malloc(2048);
-  timestr = (char*)malloc(128);
-  memset(sqlcmd,0,2048);
-
-
-  gettimeofday(tv,tz);
-  t = tv->tv_sec;
-  tm = localtime(&t);
-  strftime(timestr,128,DATE_FORMAT,tm);
+  struct ast_var_t *newvariable;
+  struct varshead *headp;
+//  struct tm tm;
+  struct timeval tv;
+  struct timezone tz;
+  char sqlcmd[8192] = "", timestart[24], timeend[24], tmp[256] = "", buf[512] = "";
+//  time_t t;
+
+  gettimeofday(&tv,&tz);
+//  t = tv.tv_sec;
+//  tm = localtime(&t);
+  strftime(timeend,sizeof(timeend),DATE_FORMAT,localtime(&(tv.tv_sec)));
+
+//  t = cdr->start.tv_sec;
+//  tm = localtime(&t);
+  strftime(timestart,sizeof(timestart),DATE_FORMAT,localtime(&(cdr->start.tv_sec)));
+  
+  headp=&cdr->parent_chan->varshead;
+                
+  AST_LIST_TRAVERSE (headp,newvariable,entries) {
+	if (strcasecmp(ast_var_name(newvariable),"tempvar"))
+	{
+		snprintf(tmp,sizeof(tmp)-1,"%s=\"%s\",",ast_var_name(newvariable),ast_var_value(newvariable));
+		strncat(buf,tmp,sizeof(buf)-strlen(buf)-1);
+	}
+  }
+  buf[sizeof(buf)-1] = 0;
+  if ( buf[strlen(buf)-1] == ',')
+	  buf[strlen(buf)-1] = 0;
   
 
   ast_log(LOG_DEBUG,"cdr_mysql: inserting a CDR record.\n");
-  sprintf(sqlcmd,"INSERT INTO cdr (calldate,clid,src,dst,dcontext,channel,dstchannel,lastapp,lastdata,duration,billsec,disposition,amaflags,accountcode) VALUES ('%s','%s','%s','%s','%s', '%s','%s','%s','%s',%i,%i,%i,%i,'%s')",timestr,cdr->clid,cdr->src, cdr->dst, cdr->dcontext,cdr->channel, cdr->dstchannel, cdr->lastapp, cdr->lastdata,cdr->duration,cdr->billsec,cdr->disposition,cdr->amaflags, cdr->accountcode);
+  sprintf(sqlcmd,"INSERT INTO cdr (callstart,callend,clid,src,dst,dcontext,channel,dstchannel,lastapp,lastdata,duration,billsec,disposition,amaflags,accountcode,vars) VALUES ('%s','%s','%s','%s','%s','%s','%s','%s','%s','%s',%i,%i,%i,%i,'%s','%s')",
+		  timestart,timeend,cdr->clid,cdr->src, cdr->dst, cdr->dcontext,cdr->channel, cdr->dstchannel, cdr->lastapp, cdr->lastdata,cdr->duration,cdr->billsec,cdr->disposition,cdr->amaflags, cdr->accountcode, buf);
   
   ast_log(LOG_DEBUG,"cdr_mysql: SQL command as follows:  %s\n",sqlcmd);
   
   if (mysql_real_query(mysql,sqlcmd,strlen(sqlcmd)))
   {
-      ast_log(LOG_ERROR,"Failed to insert into database.");
-      free(sqlcmd);
+      ast_log(LOG_ERROR,"Failed to insert into database. '%s'",sqlcmd);
       return -1;
   }
-  free(sqlcmd);
   return 0;
 }
 
@@ -93,6 +103,7 @@
   int res;
   struct ast_config *cfg;
   struct ast_variable *var;
+  char	sqlbuf[1024];
 
   char *hostname, *dbname, *dbuser, *password;
 
@@ -149,8 +160,40 @@
   } else {
     ast_log(LOG_DEBUG,"Successfully connected to MySQL database.\n");
   }
-  
+  sprintf(sqlbuf,
+	"create table if not exists cdr ("
+	"id int(10) unsigned not null auto_increment,"
+	"dstamp timestamp(14) not null,"
+	"callstart timestamp(14) not null,"
+	"callend timestamp(14) not null,"
+	"clid varchar(255),"
+	"src varchar(255),"
+	"dst varchar(255),"
+	"dcontext varchar(255),"
+	"channel varchar(255),"
+	"dstchannel varchar(255),"
+	"lastapp varchar(255),"
+	"lastdata varchar(255),"
+	"duration int(8),"
+	"billsec int(8),"
+	"disposition int(8),"
+	"amaflags int(8),"
+	"accountcode varchar(255),"
+	"vars text,"
+	"primary key (id),"
+	"key ByStartDst (callstart,disposition,dst),"
+	"key ByEndDst (callend,disposition,dst),"
+	"key ByStartSrc (callstart,disposition,src),"
+	"key ByEndSrc (callend,disposition,src),"
+	"key ByDst (dst, disposition),"
+	"key BySrc (src, disposition))");
 
+  
+  if (mysql_real_query(mysql,sqlbuf,strlen(sqlbuf)))
+  {
+      ast_log(LOG_ERROR,"Failed to create table 'cdr' '%s'.",sqlbuf);
+      return -1;
+  }
   res = ast_cdr_register(name, desc, mysql_log);
   if (res) {
     ast_log(LOG_ERROR, "Unable to register MySQL CDR handling\n");
@@ -161,6 +204,7 @@
 int reload(void)
 {
 
+  char	sqlbuf[1024];
   struct ast_config *cfg;
   struct ast_variable *var;
 
@@ -222,11 +266,41 @@
  } else {
    ast_log(LOG_DEBUG,"Successfully connected to MySQL database.\n");
  }
- 
-
+ sprintf(sqlbuf,
+	"create table if not exists cdr ("
+	"id int(10) unsigned not null auto_increment,"
+	"dstamp timestamp(14) not null,"
+	"callstart timestamp(14) not null,"
+	"callend timestamp(14) not null,"
+	"clid varchar(255),"
+	"src varchar(255),"
+	"dst varchar(255),"
+	"dcontext varchar(255),"
+	"channel varchar(255),"
+	"dstchannel varchar(255),"
+	"lastapp varchar(255),"
+	"lastdata varchar(255),"
+	"duration int(8),"
+	"billsec int(8),"
+	"disposition int(8),"
+	"amaflags int(8),"
+	"accountcode varchar(255),"
+	"vars text,"
+	"primary key (id),"
+	"key ByStartDst (callstart,disposition,dst),"
+	"key ByEndDst (callend,disposition,dst),"
+	"key ByStartSrc (callstart,disposition,src),"
+	"key ByEndSrc (callend,disposition,src),"
+	"key ByDst (dst, disposition),"
+	"key BySrc (src, disposition))");
 
-
-  return 0;
+  
+ if (mysql_real_query(mysql,sqlbuf,strlen(sqlbuf)))
+ {
+      ast_log(LOG_ERROR,"Failed to create table 'cdr'.");
+      return -1;
+ }
+ return 0;
 }
 
 int usecount(void)
diff -r -u ../asterisk.orig/cdr.c ./cdr.c
--- ../asterisk.orig/cdr.c	2003-03-07 14:25:30.000000000 +0300
+++ ./cdr.c	2003-02-25 18:56:31.000000000 +0300
@@ -175,7 +175,7 @@
 
 void ast_cdr_setapp(struct ast_cdr *cdr, char *app, char *data)
 {
-	char *chan; 
+	char *chan;
 	if (cdr) {
 		chan = strlen(cdr->channel) ? cdr->channel : "<unknown>";
 		if (cdr->posted)
@@ -222,6 +222,7 @@
 		if (strlen(cdr->channel)) 
 			ast_log(LOG_WARNING, "CDR already initialized on '%s'\n", chan); 
 		strncpy(cdr->channel, c->name, sizeof(cdr->channel) - 1);
+		cdr->parent_chan = c;
 		/* Grab source from ANI or normal Caller*ID */
 		if (c->ani)
 			strncpy(tmp, c->ani, sizeof(tmp) - 1);
diff -r -u ../asterisk.orig/channel.c ./channel.c
--- ../asterisk.orig/channel.c	2003-03-07 14:25:30.000000000 +0300
+++ ./channel.c	2003-02-25 18:56:34.000000000 +0300
@@ -1300,6 +1300,16 @@
 	
 	chan = ast_request(type, format, data);
 	if (chan) {
+		if (chan->amaflags && !chan->cdr) {
+			chan->cdr = ast_cdr_alloc();
+			if (!chan->cdr)
+				ast_log(LOG_WARNING, "Unable to create Call Detail Record\n");
+			else
+			{
+				ast_cdr_init(chan->cdr, chan);
+				ast_cdr_start(chan->cdr);
+			}
+		}
 		if (callerid)
 			ast_set_callerid(chan, callerid, 1);
 		if (!ast_call(chan, data, 0)) {
@@ -1346,7 +1356,11 @@
 	} else
 		ast_log(LOG_NOTICE, "Unable to request channel %s/%s\n", type, (char *)data);
 	if (chan && (chan->_state == AST_STATE_UP))
+	{
+		if (chan->cdr)
+			ast_cdr_answer(chan->cdr);
 		state = AST_CONTROL_ANSWER;
+	}
 	if (outstate)
 		*outstate = state;
 	return chan;
@@ -1843,7 +1857,7 @@
 			break;
 		}
 		if (c0->pvt->bridge && 
-			(c0->pvt->bridge == c1->pvt->bridge) && !nativefailed) {
+			(c0->pvt->bridge == c1->pvt->bridge) && !nativefailed && !c0->force_record && !c1->force_record) {
 				/* Looks like they share a bridge code */
 			if (option_verbose > 2) 
 				ast_verbose(VERBOSE_PREFIX_3 "Attempting native bridge of %s and %s\n", c0->name, c1->name);
@@ -1935,6 +1949,18 @@
 						ast_write(c1, f);
 					else 
 						ast_write(c0, f);
+					if (c0->force_record) {
+						if (who == c0)
+							res = ast_writestream(c0->in_stream, f);
+						else
+							res = ast_writestream(c0->out_stream, f);
+					}
+					if (c1->force_record) {
+						if (who == c1)
+							res = ast_writestream(c1->in_stream, f);
+						else
+							res = ast_writestream(c1->out_stream, f);
+					}
 				}
 			}
 			ast_frfree(f);
diff -r -u ../asterisk.orig/channels/chan_zap.c ./channels/chan_zap.c
--- ../asterisk.orig/channels/chan_zap.c	2003-03-07 14:26:11.000000000 +0300
+++ ./channels/chan_zap.c	2003-03-07 14:28:26.000000000 +0300
@@ -4724,7 +4724,7 @@
 		strncpy(tmp->callerid, callerid, sizeof(tmp->callerid)-1);
 		strncpy(tmp->mailbox, mailbox, sizeof(tmp->mailbox)-1);
 		tmp->msgstate = -1;
-		tmp->group |= cur_group;
+		tmp->group |= cur_group;
 		tmp->callgroup=cur_callergroup;
 		tmp->pickupgroup=cur_pickupgroup;
 		tmp->rxgain = rxgain;
@@ -5394,6 +5394,7 @@
 					if (ast_exists_extension(NULL, pri->pvt[chan]->context, pri->pvt[chan]->exten, 1, pri->pvt[chan]->callerid)) {
 						/* Setup law */
 						int law;
+start_on_i_exten:
 						if (e->ring.layer1 == PRI_LAYER_1_ALAW)
 							law = ZT_LAW_ALAW;
 						else
@@ -5423,10 +5424,19 @@
 						{
 							if (e->e==PRI_EVENT_RING) pri_need_more_info(pri->pri, e->ring.call, chan, 1);
 						} else {
-							if (option_verbose > 2)
-								ast_verbose(VERBOSE_PREFIX_3 "Extension '%s' in context '%s' from '%s' does not exist.  Rejecting call on channel %d, span %d\n",
-							pri->pvt[chan]->exten, pri->pvt[chan]->context, pri->pvt[chan]->callerid, chan, pri->span);
-							pri_release(pri->pri, e->ring.call, PRI_CAUSE_UNALLOCATED);
+							if (ast_exists_extension(NULL, pri->pvt[chan]->context, "i", 1, pri->pvt[chan]->callerid)) {
+								if (option_verbose > 2)
+									ast_verbose(VERBOSE_PREFIX_3 "Send invalid extension '%s' to 'i' extention in context '%s' from '%s' on channel %d, span %d\n",
+											pri->pvt[chan]->exten, pri->pvt[chan]->context, pri->pvt[chan]->callerid, chan, pri->span);
+								strncpy(pri->pvt[chan]->exten, "i", sizeof(pri->pvt[chan]->exten)-1);
+								goto start_on_i_exten;
+							}
+							else {
+								if (option_verbose > 2)
+									ast_verbose(VERBOSE_PREFIX_3 "Extension '%s' in context '%s' from '%s' does not exist.  Rejecting call on channel %d, span %d\n",
+								pri->pvt[chan]->exten, pri->pvt[chan]->context, pri->pvt[chan]->callerid, chan, pri->span);
+								pri_release(pri->pri, e->ring.call, PRI_CAUSE_UNALLOCATED);
+							}
 						}
 					}
 				} else 
diff -r -u ../asterisk.orig/include/asterisk/cdr.h ./include/asterisk/cdr.h
--- ../asterisk.orig/include/asterisk/cdr.h	2003-03-07 14:26:34.000000000 +0300
+++ ./include/asterisk/cdr.h	2003-02-25 18:57:48.000000000 +0300
@@ -33,6 +33,8 @@
 
 //! Responsible for call detail data
 struct ast_cdr {
+	/*! Channel owning CDR */
+	struct ast_channel *parent_chan;
 	/*! Caller*ID with text */
 	char clid[AST_MAX_EXTENSION];		
 	/*! Caller*ID number */
@@ -66,7 +68,7 @@
 	/*! What account number to use */
 	char accountcode[20];			
 	/*! Whether or not the record has been posted */
-	int posted;				
+	int posted;
 };
 
 typedef int (*ast_cdrbe)(struct ast_cdr *cdr);
@@ -138,6 +140,7 @@
  * \param cdr the cdr you wish to associate with the call
  * Returns nothing important
  */
+
 extern void ast_cdr_busy(struct ast_cdr *cdr);
 
 //! End a call
@@ -146,6 +149,7 @@
  * Registers the end of call time in the cdr structure.
  * Returns nothing important
  */
+ 
 extern void ast_cdr_end(struct ast_cdr *cdr);
 
 //! Post the detail record
diff -r -u ../asterisk.orig/include/asterisk/channel.h ./include/asterisk/channel.h
--- ../asterisk.orig/include/asterisk/channel.h	2003-03-07 14:26:34.000000000 +0300
+++ ./include/asterisk/channel.h	2003-02-25 18:57:49.000000000 +0300
@@ -60,6 +60,13 @@
 	/*! File descriptor for channel -- Drivers will poll on these file descriptors, so at least one must be non -1.  */
 	int fds[AST_MAX_FDS];			
 
+	/*! Force recording flag */
+	int force_record;
+	
+	/*! Poiters to the recording streams */
+	struct ast_filestream	*in_stream;
+	struct ast_filestream	*out_stream;
+	
 	/*! Default music class */
 	char musicclass[MAX_LANGUAGE];
 
diff -r -u ../asterisk.orig/pbx.c ./pbx.c
--- ../asterisk.orig/pbx.c	2003-03-07 14:25:32.000000000 +0300
+++ ./pbx.c	2003-03-07 14:28:02.000000000 +0300
@@ -1084,9 +1084,10 @@
 		return -1;
 	}
 	if (c->amaflags) {
-		if (c->cdr) {
+/*		if (c->cdr) {
 			ast_log(LOG_WARNING, "%s already has a call record??\n", c->name);
-		} else {
+		} else { */
+		if (!c->cdr) {
 			c->cdr = ast_cdr_alloc();
 			if (!c->cdr) {
 				ast_log(LOG_WARNING, "Unable to create Call Detail Record\n");
@@ -3440,8 +3441,8 @@
 {
 	int ms;
 	/* Wait for "n" seconds */
-	if (data && atoi((char *)data)) {
-		ms = atoi((char *)data) * 1000;
+	if (data && atof((char *)data)) {
+		ms = (int)(atof((char *)data) * 1000);
 		return ast_safe_sleep(chan, ms);
 	}
 	return 0;
@@ -3450,14 +3451,23 @@
 static int pbx_builtin_background(struct ast_channel *chan, void *data)
 {
 	int res;
+	char tmp[512];
+	char *options;
+	int option_noanswer = 0;
+	
+	strncpy(tmp, (char *)data, sizeof(tmp)-1);
+	strtok(tmp, "|");
+	options = strtok(NULL, "|");
+	if (options && !strcasecmp(options, "noanswer"))
+		option_noanswer = 1;
 	/* Answer if need be */
-	if (chan->_state != AST_STATE_UP)
+	if (chan->_state != AST_STATE_UP && !(options && option_noanswer))
 		if (ast_answer(chan))
 			return -1;
 	/* Stop anything playing */
 	ast_stopstream(chan);
 	/* Stream a file */
-	res = ast_streamfile(chan, (char *)data, chan->language);
+	res = ast_streamfile(chan, tmp, chan->language);
 	if (!res) {
 		res = ast_waitstream(chan, AST_DIGIT_ANY);
 		ast_stopstream(chan);
